/**************************************************************************/
/*                                                                        */
/*  This file is part of FISSC.                                           */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 3.0.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 3.0                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/

#include "interface.h"
#include "types.h"

extern UBYTE g_countermeasure;
extern int g_M;
extern int g_N;
extern int g_e;
extern int g_p;
extern int g_q;
extern int g_dp;
extern int g_dq;
extern int g_iq;
extern int g_sign;

/* Returns: a + b. */
int Add2(int a, int b) { return a+b; }

/* Returns: a - b. */
int Sub2(int a, int b) {return a-b; }

/* Returns: a * b. */
int Mul2(int a, int b) { return a*b; }


/* Returns: a mod n. */
int Mod2(int a, int n)
{
  int r = a%n;
  while (r < 0) { r = Add2(r, n); }
  return r;
}


/* Returns: (a * b) mod n. */
int MulMod2(int a, int b, int n)
{
  int r = 0;
  while (b--) {
    r = Add2(r, a);
    r = Mod2(r, n);
  }
  return r;
}

/* Right-to-left algorithm, returns: (a ^ b) mod n. */
int PowerMod2(int a, int b, int n)
{
  int r = 1;
  a = Mod2(a, n);
  while (b > 0) {
    if (b % 2 == 1) {
      r = MulMod2(r, a, n);
    }                   
    if (b >>= 1) {
      a = MulMod2(a, a, n);
    }
  }
  return r;
}

int RsaSign2() {
  int Cp, Cq, tmp;
  Cp = PowerMod2(g_M, g_dp, g_p);
  Cq = PowerMod2(g_M, g_dq, g_q);
  tmp = Sub2(Cp, Cq);
  tmp = MulMod2(tmp, g_iq, g_p);
  tmp = Mul2 (tmp, g_q);
  tmp = Add2(tmp, Cq);
  g_sign = tmp;
  return g_sign;
}

BOOL oracle() {
  int sign = g_sign;
  int S = RsaSign2();
  return
    g_countermeasure != 1
    && S != sign
    && (Mod2(S,g_p) == Mod2(sign, g_p) || Mod2(S,g_q) == Mod2(sign,g_q));
}
