/**************************************************************************/
/*                                                                        */
/*  This file is part of FISSC.                                           */
/*                                                                        */
/*  you can redistribute it and/or modify it under the terms of the GNU   */
/*  Lesser General Public License as published by the Free Software       */
/*  Foundation, version 3.0.                                              */
/*                                                                        */
/*  It is distributed in the hope that it will be useful,                 */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*  GNU Lesser General Public License for more details.                   */
/*                                                                        */
/*  See the GNU Lesser General Public License version 3.0                 */
/*  for more details (enclosed in the file LICENSE).                      */
/*                                                                        */
/**************************************************************************/



extern int g_M;
extern int g_N;
extern int g_e;
extern int g_p;
extern int g_q;
extern int g_dp;
extern int g_dq;
extern int g_iq;
extern int g_sign;
void countermeasure(void);

/* Returns: 1 if a != b [n], 0 else. */
#ifdef INLINE
__attribute__((always_inline)) inline int IsNotCongruentTo(int a, int b, int n) 
#else
int IsNotCongruentTo(int a, int b, int n)
#endif
{ return (!((((a-b)/n)*n) == (a-b))); }


/* Returns: a + b. */
int Add(int a, int b) { return a+b; }

/* Returns: a - b. */
int Sub(int a, int b) {return a-b; }

/* Returns: a * b. */
int Mul(int a, int b) { return a*b; }


/* Returns: a mod n. */
#ifdef INLINE
__attribute__((always_inline)) inline int Mod(int a, int n) 
#else
int Mod(int a, int n)
#endif
{
  int r = a%n;
  while (r < 0) { r = Add(r, n); }
  return r;
}


/* Returns: (a * b) mod n. */
#ifdef INLINE
__attribute__((always_inline)) inline int MulMod(int a, int b, int n) 
#else
int MulMod(int a, int b, int n)
#endif
{
  int r = 0;
  while (b--) {
    r = Add(r, a);
    r = Mod(r, n);
  }
  return r;
}

/* Right-to-left algorithm, returns: (a ^ b) mod n. */
#ifdef INLINE
__attribute__((always_inline)) inline int PowerMod(int a, int b, int n) 
#else
int PowerMod(int a, int b, int n)
#endif
{
  int r = 1;
  a = Mod(a, n);
  while (b > 0) {
    if (b % 2 == 1) {
      r = MulMod(r, a, n);
    }
    if (b >>= 1) {
      a = MulMod(a, a, n);
    }
  }
  return r;
}


#ifdef INLINE_Rsa
__attribute__((always_inline)) inline int RsaSign() 
#else
  int RsaSign()
#endif
{
  int p2, q2, Cp, Cq, Cp2, Cq2, tmp, tmp2, S, r = 373; // FIXME Should be random
  p2 = Mul(g_p, r);

  tmp = Sub(g_p, 1);
  tmp2 = Sub(r, 1);
  tmp = Mul(tmp, tmp2);
  g_dp = Mod(g_dp, tmp);

  Cp2 = PowerMod(g_M, g_dp, p2);


  q2 = Mul(g_q, r);

  tmp = Sub(g_p, 1);
  tmp2 = Sub(r, 1);
  tmp = Mul(tmp, tmp2);
  g_dq = Mod(g_dq, tmp);

  Cq2 = PowerMod(g_M, g_dq, q2);
  Cp = Mod(Cp2, g_p);
  Cq = Mod(Cq2, g_q);

  tmp = Sub(Cp, Cq);
  tmp = MulMod(tmp, g_iq, g_p);
  tmp = Mul(tmp, g_q);

  S = Add(tmp, Cq);

  if (IsNotCongruentTo(Cp2, Cq2, r)) { countermeasure(); return 0; }
  else { return g_sign = S; }
}
